# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Tuple

import av
import numpy as np
from av.codec.codec import UnknownCodecError

from audiolab.av import from_ndarray
from audiolab.av.format import dtype_formats
from audiolab.av.frame import clip
from audiolab.av.layout import standard_channel_layouts
from audiolab.av.typing import ContainerFormat, Dtype
from audiolab.writer.backend.backend import Backend


class PyAV(Backend):
    def __init__(self, file: Any, sample_rate: int, dtype: Optional[Dtype] = None, format: ContainerFormat = "WAV"):
        super().__init__(file, sample_rate, dtype, format)
        self.container = av.open(self.file, "w", format=self.format)
        self.num_channels = None
        self.stream = None

    def open(self):
        kwargs = {"layout": standard_channel_layouts[self.num_channels][0]}
        audio_codec, audio_format = self.guess_codec_format()
        if audio_format is not None:
            kwargs["format"] = audio_format
        self.stream = self.container.add_stream(audio_codec, self.sample_rate, **kwargs)

    def guess_codec_format(self) -> Tuple[str, str]:
        default_codec = self.container.default_audio_codec
        if self.dtype is None:
            return default_codec, None
        else:
            dtype_format = dtype_formats[self.dtype]
            for audio_format in av.Codec(default_codec, "w").audio_formats:
                if audio_format.name.startswith(dtype_format):
                    return default_codec, audio_format.name

            supported_codecs = self.container.supported_codecs
            codecs = sorted(supported_codecs, key=lambda x: (not x.startswith("pcm_") or x.endswith("law"), x))
            for codec in codecs:
                try:
                    audio_formats = av.Codec(codec, "w").audio_formats
                    if audio_formats is None:
                        continue
                    for audio_format in audio_formats:
                        if audio_format.name.startswith(dtype_format):
                            return codec, audio_format.name
                except UnknownCodecError:
                    pass

    def write(self, frame: np.ndarray):
        if self.dtype is None:
            self.dtype = frame.dtype
        frame = np.atleast_2d(clip(frame, self.dtype))
        if self.num_channels is None:
            self.num_channels = frame.shape[0]
        if self.stream is None:
            self.open()
        frame = from_ndarray(frame, self.stream.format.name, self.stream.layout, self.stream.rate)
        for packet in self.stream.encode(frame):
            self.container.mux(packet)

    def close(self):
        if not self.is_closed:
            try:
                for packet in self.stream.encode():
                    self.container.mux(packet)
            except ValueError:
                pass
            self.container.close()
            super().close()



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import lru_cache
from typing import Dict, Iterator, Literal, Optional, Set

import av
import numpy as np
from av import Codec, codecs_available
from av.codec.codec import UnknownCodecError

from audiolab.av import typing
from audiolab.av.utils import get_template

"""
$ ffmpeg -sample_fmts
"""
format_dtypes = {
    "dbl": "f8",
    "dblp": "f8",
    "flt": "f4",
    "fltp": "f4",
    "s16": "i2",
    "s16p": "i2",
    "s32": "i4",
    "s32p": "i4",
    "s64": "i8",
    "s64p": "i8",
    "u8": "u1",
    "u8p": "u1",
}
dtype_formats = {np.dtype(dtype): name for name, dtype in format_dtypes.items() if not name.endswith("p")}
audio_formats: Dict[str, av.AudioFormat] = {name: av.AudioFormat(name) for name in format_dtypes.keys()}
AudioFormat = typing.AudioFormatEnum("AudioFormat", audio_formats)


@lru_cache(maxsize=None)
def get_codecs(format: typing.AudioFormat, mode: Literal["r", "w"] = "r") -> Set[str]:
    codecs = set()
    if isinstance(format, av.AudioFormat):
        format = format.name
    for codec in codecs_available:
        try:
            codec = Codec(codec, mode)
            formats = codec.audio_formats
            if codec.type != "audio" or formats is None:
                continue
            if format in set(format.name for format in formats):
                codecs.add(codec.name)
        except UnknownCodecError:
            pass
    return codecs


@lru_cache(maxsize=None)
def get_dtype(format: typing.AudioFormat) -> np.dtype:
    if isinstance(format, av.AudioFormat):
        format = format.name
    return np.dtype(format_dtypes[format])


def get_format(
    dtype: typing.Dtype,
    is_planar: Optional[bool] = None,
    available_formats: Optional[Iterator[typing.AudioFormat]] = None,
) -> av.AudioFormat:
    if isinstance(dtype, str) and dtype not in format_dtypes or isinstance(dtype, type):
        dtype = np.dtype(dtype)
    if isinstance(dtype, np.dtype):
        dtype = dtype_formats[dtype]
        if is_planar is not None:
            dtype = dtype + ("p" if is_planar else "")
        else:
            assert available_formats is not None
            available_formats = [
                format.name if isinstance(format, typing.AudioFormat) else format for format in available_formats
            ]
            if dtype not in available_formats:
                dtype = dtype.rstrip("p") if dtype.endswith("p") else dtype + "p"
    return AudioFormat[dtype].value


template = get_template("format")
for name, format in audio_formats.items():
    decodecs = get_codecs(name, "r")
    encodecs = get_codecs(name, "w")
    dtype = get_dtype(name)
    getattr(AudioFormat, name).__doc__ = template.render(
        format=format, decodecs=decodecs, encodecs=encodecs, dtype=dtype
    )



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property, partial
from typing import Any, Iterator, List, Optional

from audiolab.av import aformat, load_url
from audiolab.av.frame import pad
from audiolab.av.graph import Graph
from audiolab.av.typing import UINT32_MAX, AudioFrame, Dtype, Filter, Seconds
from audiolab.reader.backend import pyav, soundfile
from audiolab.reader.info import Info


class Reader(Info):
    def __init__(
        self,
        file: Any,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        filters: Optional[List[Filter]] = None,
        dtype: Optional[Dtype] = None,
        rate: Optional[int] = None,
        to_mono: bool = False,
        frame_size: Optional[int] = None,
        cache_url: bool = False,
        always_2d: bool = True,
        fill_value: Optional[float] = None,
    ):
        """
        Create a Reader object.

        Args:
            file: The audio file, audio url, path to audio file, bytes of audio data, etc.
            offset: The offset of the audio to load.
            duration: The duration of the audio to load.
            filters: The filters to apply to the audio.
            dtype: The data type of the audio frames.
            rate: The sample rate of the audio frames.
            to_mono: Whether to convert the audio frames to mono.
            frame_size: The frame size of the audio frames.
            cache_url: Whether to cache the audio file.
            always_2d: Whether to return 2d ndarrays even if the audio frame is mono.
            fill_value: The fill value to pad the audio to the frame size.
        """
        if isinstance(file, str) and "://" in file:
            if cache_url:
                file = load_url(file, cache=True)
            elif offset == 0 and duration is None:
                file = load_url(file, cache=False)

        self.filters = [] if filters is None else filters
        if rate is None and not to_mono and len(self.filters) == 0:
            if dtype is None:
                super().__init__(file, frame_size)
            else:
                super().__init__(file, frame_size, backends=[soundfile, pyav])
                if isinstance(self.backend, soundfile):
                    self.backend.read = partial(self.backend.read, dtype=dtype)
        else:
            super().__init__(file, frame_size, backends=[pyav])

        self.graph = None
        if not self.is_passthrough(dtype, rate, to_mono):
            self.filters.append(aformat(dtype, rate=rate, to_mono=to_mono))
            graph = Graph(
                rate=self.rate,
                dtype=self.dtype,
                is_planar=self.backend.is_planar,
                channels=self.num_channels,
                filters=self.filters,
                frame_size=self.frame_size,
            )
            if isinstance(self.backend, pyav):
                self.backend.graph = graph
            else:
                self.graph = graph
        self.offset = offset
        self._duration = duration
        self.always_2d = always_2d
        self.fill_value = fill_value

    @cached_property
    def frame_size(self) -> int:
        return self.backend.frame_size

    def __iter__(self) -> Iterator[AudioFrame]:
        for frame in self.backend.load_audio(self.offset, self._duration):
            if self.graph is None:
                rate = self.rate
                if isinstance(self.backend, pyav):
                    frame, rate = frame
                if self.fill_value is not None:
                    frame = pad(frame, self.frame_size, self.fill_value)
                yield frame if self.always_2d else frame.squeeze(), rate
            else:
                self.graph.push(frame)
                yield from self.pull()
        if self.graph is not None:
            yield from self.pull(partial=True)

    def is_passthrough(self, dtype: Optional[Dtype] = None, rate: Optional[int] = None, to_mono: bool = False) -> bool:
        to_mono = to_mono and self.num_channels > 1
        passthrough = self.dtype == dtype and self.rate == rate and not to_mono
        return passthrough and self.frame_size >= UINT32_MAX and len(self.filters) == 0

    def pull(self, partial: bool = False) -> AudioFrame:
        for frame in self.graph.pull(partial=partial):
            frame, rate = frame
            if self.fill_value is not None:
                frame = pad(frame, self.frame_size, self.fill_value)
            yield frame if self.always_2d else frame.squeeze(), rate



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import cached_property
from io import BytesIO
from typing import Any, Iterator, Optional

import numpy as np

from audiolab.av import standard_channel_layouts
from audiolab.av.typing import UINT32_MAX, Seconds


class Backend:
    def __init__(self, file: Any, frame_size: Optional[int] = None, forced_decoding: bool = False):
        self.file = file
        self.frame_size = UINT32_MAX if frame_size is None else min(frame_size, UINT32_MAX)
        self.forced_decoding = forced_decoding

    @cached_property
    def bit_rate(self) -> Optional[int]:
        bit_rate = None
        if self.size is not None:
            if self.duration is not None and self.duration > 0:
                bit_rate = self.size * 8 / self.duration
        return bit_rate

    @cached_property
    def is_planar(self) -> bool:
        return False

    @cached_property
    def layout(self) -> str:
        layouts = standard_channel_layouts[self.num_channels]
        return layouts[0]

    @cached_property
    def metadata(self) -> dict:
        return {}

    @cached_property
    def name(self) -> str:
        return "<none>" if isinstance(self.file, BytesIO) else self.file

    @cached_property
    def size(self) -> Optional[int]:
        if isinstance(self.file, str):
            if os.path.exists(self.file):
                return os.stat(self.file).st_size
        elif isinstance(self.file, BytesIO):
            return len(self.file.getbuffer())
        return None

    def load_audio(self, offset: Seconds = 0, duration: Optional[Seconds] = None) -> Iterator[np.ndarray]:
        self.seek(int(offset * self.sample_rate))
        frames = UINT32_MAX if duration is None else int(duration * self.sample_rate)
        while frames > 0:
            frame_size = min(frames, self.frame_size)
            ndarray = self.read(frame_size)
            if ndarray is None:
                break
            frames -= ndarray.shape[1]
            yield ndarray



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property
from typing import Any, Iterator, Optional

import av
import numpy as np
from av import time_base
from av.codec import Codec

from audiolab.av import split_audio_frame
from audiolab.av.format import get_dtype
from audiolab.av.graph import Graph
from audiolab.av.typing import UINT32_MAX, AudioFrame, Seconds
from audiolab.reader.backend.backend import Backend


class PyAV(Backend):
    def __init__(self, file: Any, frame_size: Optional[int] = None, forced_decoding: bool = False):
        super().__init__(file, frame_size, forced_decoding)
        self.container = av.open(file, metadata_encoding="latin1")
        self.stream = self.container.streams.audio[0]
        self.graph = Graph(
            rate=self.sample_rate,
            dtype=self.dtype,
            is_planar=self.is_planar,
            channels=self.num_channels,
            frame_size=self.frame_size,
        )

    @cached_property
    def bits_per_sample(self) -> int:
        return self.stream.format.bits

    @cached_property
    def bit_rate(self) -> Optional[int]:
        bit_rate = None
        if self.stream.bit_rate is not None:
            bit_rate = self.stream.bit_rate
        elif self.container.bit_rate is not None:
            bit_rate = self.container.bit_rate
        if bit_rate in (0, None):
            bit_rate = super().bit_rate
        return bit_rate

    @cached_property
    def codec(self) -> Codec:
        return self.stream.codec.long_name

    @cached_property
    def dtype(self) -> np.dtype:
        return get_dtype(self.stream.format)

    @cached_property
    def format(self) -> str:
        return self.container.format.name

    @cached_property
    def duration(self) -> Optional[Seconds]:
        if self.forced_decoding:
            num_frames = 0
            for frame in self.container.decode(self.stream):
                num_frames += frame.samples
            duration = num_frames / self.stream.rate
        else:
            duration = None
            if self.stream.duration is not None:
                duration = self.stream.duration * self.stream.time_base
            elif self.container.duration is not None:
                duration = self.container.duration / time_base
        return None if duration is None else Seconds(duration)

    @cached_property
    def is_planar(self) -> bool:
        return self.stream.format.is_planar

    @cached_property
    def name(self) -> str:
        return self.container.name

    @cached_property
    def num_channels(self) -> int:
        return self.stream.channels

    @cached_property
    def num_frames(self) -> Optional[int]:
        if self.duration is None:
            return None
        return int(self.duration * self.stream.rate)

    @cached_property
    def metadata(self) -> dict:
        return {**self.container.metadata, **self.stream.metadata}

    @cached_property
    def sample_rate(self) -> int:
        return self.stream.sample_rate

    @cached_property
    def size(self) -> Optional[int]:
        size = super().size
        if size is None:
            size = self.container.size
        return size

    @cached_property
    def seekable(self) -> bool:
        # https://github.com/FFmpeg/FFmpeg/blob/master/libavcodec/avcodec.h#L1041-L1051
        # Each submitted frame except the last must contain exactly frame_size samples per channel.
        # May be 0 when the codec has AV_CODEC_CAP_VARIABLE_FRAME_SIZE set, then the frame size is not restricted.
        return self.stream.codec_context.frame_size in (0, 1)

    def load_audio(self, offset: Seconds = 0, duration: Optional[Seconds] = None) -> Iterator[AudioFrame]:
        self.seek(int(offset / self.stream.time_base))
        frames = UINT32_MAX if duration is None else int(duration * self.sample_rate)
        while frames > 0:
            frame = self.read()
            if frame is None:
                break
            frame, _ = split_audio_frame(frame, frames)
            frames -= frame.samples
            self.graph.push(frame)
            yield from self.graph.pull()
        yield from self.graph.pull(partial=True)

    def read(self) -> Optional[AudioFrame]:
        try:
            return next(self.container.decode(self.stream))
        except StopIteration:
            return None

    def seek(self, offset: int):
        self.container.seek(offset, any_frame=True, stream=self.stream)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

import numpy as np
import soundfile as sf

from audiolab.av.typing import Dtype
from audiolab.writer.backend import pyav, soundfile


class Writer:
    def __init__(self, file: Any, rate: int, dtype: Optional[Dtype] = None, format: str = "WAV"):
        backend = soundfile if format.upper() in sf.available_formats() else pyav
        self.backend = backend(file, rate, dtype, format)

    def write(self, frame: np.ndarray):
        self.backend.write(frame)

    def close(self):
        self.backend.close()



# Copyright (c) 2021 Piotr Żelasko
# From https://github.com/lhotse-speech/lhotse/blob/master/lhotse/caching.py
#      https://github.com/lhotse-speech/lhotse/blob/master/lhotse/utils.py
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from threading import Lock
from typing import Callable, Dict, Optional

from smart_open import open as sm_open

from audiolab.av.utils import get_logger

logger = get_logger(__name__)


class AudioCache:
    """
    Cache of 'bytes' objects with audio data.
    It is used to cache the "command" type audio inputs.

    The cache size is limited to max 100 elements and 500MB of audio.

    A global dict `__cache_dict` (static member variable of class AudioCache)
    is holding the wavs as 'bytes' arrays.
    The key is the 'source' identifier (i.e. the command for loading the data).

    Thread-safety is ensured by a threading.Lock guard.
    """

    max_cache_memory: int = 500 * 1e6  # 500 MB
    max_cache_elements: int = 100  # 100 audio files

    __cache_dict: Dict[str, bytes] = {}
    __cache_memory: int = 0
    __lock: Lock = Lock()

    @classmethod
    def try_cache(cls, key: str) -> Optional[bytes]:
        """
        Test if 'key' is in the chache. If yes return the bytes array,
        otherwise return None.
        """

        with cls.__lock:
            if key in cls.__cache_dict:
                return cls.__cache_dict[key]
            else:
                return None

    @classmethod
    def add_to_cache(cls, key: str, value: bytes):
        """
        Add the new (key,value) pair to cache.
        Possibly free some elements before adding the new pair.
        The oldest elements are removed first.
        """

        if len(value) > cls.max_cache_memory:
            return

        with cls.__lock:
            # limit cache elements and memory
            while (
                len(cls.__cache_dict) >= cls.max_cache_elements
                or len(value) + cls.__cache_memory > cls.max_cache_memory
            ):
                # remove oldest elements from cache
                # (dict pairs are sorted according to insertion order)
                removed_key = next(iter(cls.__cache_dict))
                removed_value = cls.__cache_dict.pop(removed_key)
                cls.__cache_memory -= len(removed_value)

            # store the new (key,value) pair
            cls.__cache_dict[key] = value
            cls.__cache_memory += len(value)

    @property
    def cache_memory(cls) -> int:
        """
        Return size of AudioCache values in bytes.
        """
        return cls.__cache_memory

    @classmethod
    def clear_cache(cls) -> None:
        """
        Clear the cache, remove the data.
        """
        with cls.__lock:
            cls.__cache_dict.clear()
            cls.__cache_memory = 0


class SmartOpen:
    """Wrapper class around smart_open.open method

    The smart_open.open attributes are cached as classed attributes - they play the role of singleton pattern.

    The SmartOpen.setup method is intended for initial setup.
    It imports the `open` method from the optional `smart_open` Python package,
    and sets the parameters which are shared between all calls of the `smart_open.open` method.

    If you do not call the setup method it is called automatically in SmartOpen.open with the provided parameters.

    The example demonstrates that instantiating S3 `session.client` once,
    instead using the defaults and leaving the smart_open creating it every time
    has dramatic performance benefits.
    """

    transport_params: Optional[Dict] = None
    smart_open: Optional[Callable] = None

    @classmethod
    def setup(cls, transport_params: Optional[dict] = None):
        if cls.transport_params is not None and cls.transport_params != transport_params:
            logger.warning(
                "SmartOpen.setup second call overwrites existing transport_params with new version\t\n%s\t\nvs\t\n%s",
                cls.transport_params,
                transport_params,
            )
        cls.transport_params = transport_params
        cls.smart_open = sm_open

    @classmethod
    def open(cls, uri, mode="rb", transport_params=None, **kwargs):
        if cls.smart_open is None:
            cls.setup(transport_params=transport_params)
        transport_params = transport_params if transport_params else cls.transport_params
        return cls.smart_open(
            uri,
            mode=mode,
            transport_params=transport_params,
            **kwargs,
        )


def load_url(url: str, cache: bool = False) -> BytesIO:
    """
    Load an audio file from a URL.

    Args:
        url (str): The URL of the audio file.
        cache (bool): Whether to cache the audio file.
    Returns:
        The audio bytes.
    """
    audio_bytes = AudioCache.try_cache(url) if cache else None
    if audio_bytes is None:
        with SmartOpen.open(url, "rb") as f:
            audio_bytes = f.read()
        if cache:
            AudioCache.add_to_cache(url, audio_bytes)
    return BytesIO(audio_bytes)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import wave
from functools import cached_property
from typing import Any, Optional

import numpy as np
from av.codec import Codec

from audiolab.av.typing import Seconds
from audiolab.reader.backend.backend import Backend

_bits_to_codec = {8: "pcm_u8le", 16: "pcm_s16le", 24: "pcm_s32le", 32: "pcm_s32le"}
_bits_to_dtype = {8: np.uint8, 16: np.int16, 24: np.int32, 32: np.int32}


class Wave(Backend):
    def __init__(self, file: Any, frame_size: Optional[int] = None, forced_decoding: bool = False):
        super().__init__(file, frame_size, forced_decoding)
        self.wave = wave.open(file)

    @cached_property
    def bits_per_sample(self) -> int:
        return self.wave.getsampwidth() * 8

    @cached_property
    def codec(self) -> str:
        return Codec(_bits_to_codec[self.bits_per_sample]).long_name

    @cached_property
    def duration(self) -> Optional[Seconds]:
        if self.num_frames is None:
            return None
        return Seconds(self.num_frames / self.sample_rate)

    @cached_property
    def dtype(self) -> np.dtype:
        return _bits_to_dtype[self.bits_per_sample]

    @cached_property
    def format(self) -> str:
        return "WAV"

    @cached_property
    def num_channels(self) -> int:
        return self.wave.getnchannels()

    @cached_property
    def num_frames(self) -> Optional[int]:
        if self.forced_decoding:
            num_frames = self.read(np.iinfo(np.int32).max).shape[0]
            self.wave.rewind()
        else:
            num_frames = self.wave.getnframes()
            if num_frames >= np.iinfo(np.int32).max:
                num_frames = None
        return num_frames

    @cached_property
    def sample_rate(self) -> int:
        return self.wave.getframerate()

    @cached_property
    def seekable(self) -> bool:
        return True

    def frombuffer(self, buffer: bytes) -> np.ndarray:
        if self.bits_per_sample == 24:
            frames = np.frombuffer(buffer, np.uint8)
            frames = (
                (frames[2::3].astype(np.int32) << 16)
                | (frames[1::3].astype(np.int32) << 8)
                | frames[0::3].astype(np.int32)
            )
            frames[frames > 0x7FFFFF] -= 0x1000000
        else:
            frames = np.frombuffer(buffer, self.dtype)
        return frames.reshape(-1, self.num_channels).T

    def read(self, nframes: int) -> Optional[np.ndarray]:
        buffer = self.wave.readframes(nframes)
        return self.frombuffer(buffer) if len(buffer) > 0 else None

    def seek(self, offset: int):
        self.wave.setpos(offset)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from importlib.resources import files

import numpy as np
from jinja2 import Environment, FileSystemLoader
from numpy.random import randint, uniform

loader = FileSystemLoader(files("audiolab.av").joinpath("templates"))


def generate_ndarray(nb_channels: int, samples: int, dtype: np.dtype, always_2d: bool = True) -> np.ndarray:
    if np.dtype(dtype).kind in ("i", "u"):
        ndarray = randint(np.iinfo(dtype).min, np.iinfo(dtype).max, size=(nb_channels, samples), dtype=dtype)
    else:
        ndarray = uniform(-1, 1, size=(nb_channels, samples)).astype(dtype)
    return ndarray if always_2d else ndarray.squeeze()


def get_template(name: str) -> str:
    return Environment(loader=loader).get_template(f"{name}.txt")


def get_logger(name, level=logging.INFO):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if not logger.handlers:
        logger.propagate = False
        handler = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property
from typing import Any, Optional

import numpy as np
import soundfile as sf

from audiolab.av.typing import Dtype, Seconds
from audiolab.reader.backend.backend import Backend

_subtype_to_bits = {
    "PCM_S8": 8,
    "PCM_U8": 8,
    "PCM_16": 16,
    "PCM_24": 24,
    "PCM_32": 32,
    "FLOAT": 32,
    "DOUBLE": 64,
    "ULAW": 8,
    "ALAW": 8,
    "DWVW_12": 12,
    "DWVW_16": 16,
    "DWVW_24": 24,
    "DPCM_8": 8,
    "DPCM_16": 16,
    "ALAC_16": 16,
    "ALAC_20": 20,
    "ALAC_24": 24,
    "ALAC_32": 32,
}

_subtype_to_dtype = {
    "PCM_S8": np.int8,
    "PCM_U8": np.uint8,
    "PCM_16": np.int16,
    "PCM_24": np.int32,
    "PCM_32": np.int32,
    "FLOAT": np.float32,
    "DOUBLE": np.float64,
}


class SoundFile(Backend):
    def __init__(self, file: Any, frame_size: Optional[int] = None, forced_decoding: bool = False):
        super().__init__(file, frame_size, forced_decoding)
        self.sf = sf.SoundFile(file)

    @cached_property
    def bits_per_sample(self) -> Optional[int]:
        return _subtype_to_bits.get(self.sf.subtype, None)

    @cached_property
    def codec(self) -> str:
        return sf.available_subtypes()[self.sf.subtype]

    @cached_property
    def duration(self) -> Optional[Seconds]:
        if self.num_frames is None:
            return None
        return Seconds(self.num_frames / self.sample_rate)

    @cached_property
    def dtype(self) -> np.dtype:
        return _subtype_to_dtype.get(self.sf.subtype, np.float32)

    @cached_property
    def format(self) -> str:
        return self.sf.format

    @cached_property
    def num_channels(self) -> int:
        return self.sf.channels

    @cached_property
    def num_frames(self) -> Optional[int]:
        if self.forced_decoding:
            num_frames = 0
            pos = self.sf.tell()
            try:
                frames = self.sf.read()
                num_frames = frames.shape[0]
            except sf.LibsndfileError:
                self.sf = sf.SoundFile(self.file)
            self.seek(pos)
        else:
            num_frames = self.sf.frames
            if num_frames >= np.iinfo(np.int32).max:
                num_frames = None
        return num_frames

    @cached_property
    def metadata(self) -> dict:
        return self.sf.copy_metadata()

    @cached_property
    def sample_rate(self) -> int:
        return self.sf.samplerate

    @cached_property
    def seekable(self) -> bool:
        return self.sf.seekable()

    def read(self, nframes: int, dtype: Optional[Dtype] = None) -> Optional[np.ndarray]:
        if dtype is None:
            dtype = self.dtype
        frames = self.sf.read(nframes, dtype=dtype)
        return np.atleast_2d(frames.T) if frames.shape[0] > 0 else None

    def seek(self, offset: int):
        self.sf.seek(offset)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property
from io import BytesIO
from typing import Any, List, Optional, Union

import numpy as np
from av.codec import Codec
from humanize import naturalsize

from audiolab.av.typing import Seconds
from audiolab.av.utils import get_template
from audiolab.reader.backend import Backend, pyav, soundfile


class Info:
    def __init__(
        self,
        file: Any,
        frame_size: Optional[int] = None,
        forced_decoding: bool = False,
        backends: Optional[List[Backend]] = None,
    ):
        self.file = file
        if backends is None:
            backends = [soundfile, pyav]

        for idx, backend in enumerate(backends):
            pos = file.tell() if isinstance(file, BytesIO) else 0
            try:
                self.backend = backend(file, frame_size, forced_decoding)
                break
            except Exception as e:
                if isinstance(file, BytesIO):
                    file.seek(pos)
                if idx == len(backends) - 1:
                    raise e

    @cached_property
    def bits_per_sample(self) -> int:
        return self.backend.bits_per_sample

    @property
    def bit_rate(self) -> Optional[int]:
        return self.backend.bit_rate

    @cached_property
    def codec(self) -> Union[Codec, str]:
        return self.backend.codec

    @cached_property
    def duration(self) -> Optional[Seconds]:
        return self.backend.duration

    @cached_property
    def dtype(self) -> np.dtype:
        return self.backend.dtype

    @cached_property
    def format(self) -> str:
        return self.backend.format

    @cached_property
    def layout(self) -> str:
        return self.backend.layout

    @cached_property
    def name(self) -> str:
        return self.backend.name

    @property
    def num_channels(self) -> int:
        return self.backend.num_channels

    @property
    def num_frames(self) -> int:
        return self.backend.num_frames

    @property
    def metadata(self) -> int:
        return self.backend.metadata

    @property
    def sample_rate(self) -> int:
        return self.backend.sample_rate

    @property
    def size(self) -> int:
        return self.backend.size

    @property
    def cdda_sectors(self) -> Optional[float]:
        if self.duration is None:
            return None
        return round(self.duration * 75, 2)

    @property
    def channels(self) -> int:
        return self.num_channels

    @property
    def num_samples(self) -> int:
        # Number of audio samples (per channel).
        return self.num_frames

    @property
    def rate(self) -> int:
        return self.sample_rate

    @property
    def samplerate(self) -> int:
        return self.sample_rate

    @property
    def samples(self) -> int:
        return self.backend.num_frames

    @property
    def precision(self) -> int:
        return self.bits_per_sample

    @staticmethod
    def rstrip_zeros(s: Optional[Union[int, float, str]]) -> str:
        if s is None:
            return "N/A"
        if not isinstance(s, str):
            s = str(s)
        return " ".join(x.rstrip("0").rstrip(".") for x in s.split())

    @staticmethod
    def format_bit_rate(bit_rate: Union[int, None]) -> str:
        if bit_rate is None or bit_rate <= 0:
            return "N/A"
        bit_rate = naturalsize(bit_rate).rstrip("B")
        return Info.rstrip_zeros(bit_rate) + "bps"

    @staticmethod
    def format_duration(duration: Union[Seconds, None]) -> str:
        if duration is None:
            return "N/A"
        hours, rest = divmod(duration, 3600)
        minutes, seconds = divmod(rest, 60)
        return f"{int(hours):02d}:{int(minutes):02d}:{seconds:06.3f}"

    @staticmethod
    def format_name(name: str, format: str) -> str:
        if name.upper().endswith(format.upper()):
            return f"'{name}'"
        if name in ("<none>", "<stdin>"):
            return f"{name} ({format})"
        return f"'{name}' ({format})"

    @staticmethod
    def format_size(size: int) -> str:
        if size not in (-1, -38, -78, None):
            size = naturalsize(size)
        return Info.rstrip_zeros(size)

    def __str__(self):
        return get_template("info").render(
            name=Info.format_name(self.name, self.format),
            channels=self.channels,
            rate=self.rate,
            precision=self.precision,
            duration=Info.format_duration(self.duration),
            samples="N/A" if self.samples is None else self.samples,
            cdda_sectors=Info.rstrip_zeros(self.cdda_sectors),
            size=Info.format_size(self.size),
            bit_rate=Info.format_bit_rate(self.bit_rate),
            codec=self.codec,
            metadata=self.metadata,
        )



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Iterator, List, Optional, Union

from audiolab.av import aformat
from audiolab.av.graph import Graph
from audiolab.av.typing import UINT32_MAX, AudioFrame
from audiolab.reader.backend import Backend
from audiolab.reader.info import Info
from audiolab.reader.reader import Reader
from audiolab.reader.stream_reader import StreamReader


def info(file: Any, forced_decoding: bool = False, backends: Optional[List[Backend]] = None) -> Info:
    """
    Get the information of an audio file.

    Args:
        file: The input audio file, audio url, path to audio file, bytes of audio data, etc.
        forced_decoding: Whether to forced decoding the audio file to get the duration.
        backends: The list of backends to use to get the information.
    Returns:
        The information of the audio file.
    """
    return Info(file, forced_decoding=forced_decoding, backends=backends)


def load_audio(file: Any, **kwargs) -> Union[Iterator[AudioFrame], AudioFrame]:
    reader = Reader(file, **kwargs)
    if reader.frame_size < UINT32_MAX:
        return iter(reader)
    else:
        return next(iter(reader))


__all__ = ["Graph", "Reader", "StreamReader", "aformat", "load_audio"]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from typing import Any

import click

import audiolab
from audiolab.reader.backend import pyav
from audiolab.reader.info import Info


@click.command()
@click.argument("audio-files", nargs=-1)
@click.option(
    "-f",
    "--forced-decoding",
    is_flag=True,
    help="Forced decoding the audio file to get the duration",
)
@click.option("-t", "--show-file-type", is_flag=True, help="Show detected file-type")
@click.option("-r", "--show-sample-rate", is_flag=True, help="Show sample-rate")
@click.option("-c", "--show-channels", is_flag=True, help="Show number of channels")
@click.option(
    "-s",
    "--show-samples",
    is_flag=True,
    help="Show number of samples (N/A if unavailable)",
)
@click.option(
    "-d",
    "--show-duration-hms",
    is_flag=True,
    help="Show duration in hours, minutes and seconds (N/A if unavailable)",
)
@click.option(
    "-D",
    "--show-duration-seconds",
    is_flag=True,
    help="Show duration in seconds (N/A if unavailable)",
)
@click.option(
    "-b",
    "--show-bits-per-sample",
    is_flag=True,
    help="Show number of bits per sample (N/A if not applicable)",
)
@click.option(
    "-B",
    "--show-bitrate",
    is_flag=True,
    help="Show the bitrate averaged over the whole file (N/A if unavailable)",
)
@click.option(
    "-p",
    "--show-precision",
    is_flag=True,
    help="Show estimated sample precision in bits",
)
@click.option("-e", "--show-encoding", is_flag=True, help="Show the name of the audio encoding")
@click.option(
    "-a",
    "--show-comments",
    is_flag=True,
    help="Show file comments (annotations) if available",
)
def main(
    audio_files: Any,
    forced_decoding: bool = False,
    show_file_type: bool = False,
    show_sample_rate: bool = False,
    show_channels: bool = False,
    show_samples: bool = False,
    show_duration_hms: bool = False,
    show_duration_seconds: bool = False,
    show_bits_per_sample: bool = False,
    show_bitrate: bool = False,
    show_precision: bool = False,
    show_encoding: bool = False,
    show_comments: bool = False,
):
    """
    Print the information of audio files.

    Args:

        AUDIO_FILES: The audio files, audio urls, paths to audio files, or stdin.
    """
    # If no files are provided, use stdin
    if not audio_files:
        # Create a file object for stdin
        #   cat audio.wav | audi
        #   audi < audio.wav
        stdin_file = click.File(mode="rb").convert("-", None, None)
        bytesio = BytesIO(stdin_file.read())
        audio_files = [bytesio]

    # Initialize total duration and show_any flag
    total_duration = 0.0
    show_any = any(
        [
            show_file_type,
            show_sample_rate,
            show_channels,
            show_samples,
            show_duration_hms,
            show_duration_seconds,
            show_bits_per_sample,
            show_bitrate,
            show_precision,
            show_encoding,
            show_comments,
        ]
    )

    # Process each audio file
    for audio_file in audio_files:
        # ffmpeg -i audio.flac -f wav - | > audio.wav
        info = audiolab.info(audio_file, forced_decoding, backends=[pyav])
        # If no specific options are selected, show all information (default behavior)
        if not show_any:
            print(info)
            # Accumulate total duration
            total_duration += info.duration or 0.0

        # Display information based on selected options
        if show_file_type:
            print(info.container.format.name)
        if show_sample_rate:
            print(info.sample_rate)
        if show_channels:
            print(info.channels)
        if show_samples:
            print(info.num_samples or 0)
        if show_duration_hms:
            print(Info.format_duration(info.duration))
        if show_duration_seconds:
            print(info.duration or 0)
        if show_bits_per_sample:
            print(info.precision)
        if show_bitrate:
            print(Info.format_bit_rate(info.bit_rate))
        if show_precision:
            print(info.precision)
        if show_encoding:
            print(info.codec.long_name)
        if show_comments:
            if info.metadata:
                for key, value in info.metadata.items():
                    print(f"{key}: {value}")

    # Print total duration if any files were processed and any options were selected
    if len(audio_files) > 1 and not show_any:
        print(f"\nTotal duration of {len(audio_files)} files: {Info.format_duration(total_duration)}")



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import av
import numpy as np

from audiolab.av import filter
from audiolab.av.codec import Decodec, Encodec, canonical_names, decodecs, encodecs
from audiolab.av.container import ContainerFormat, container_formats, extension_formats
from audiolab.av.format import AudioFormat, audio_formats, get_codecs, get_dtype, get_format
from audiolab.av.frame import clip, from_ndarray, split_audio_frame, to_ndarray
from audiolab.av.graph import Graph
from audiolab.av.layout import AudioLayout, audio_layouts, standard_channel_layouts
from audiolab.av.lhotse import AudioCache, load_url


def aformat(
    dtype: Optional[Union[str, type, np.dtype]] = None,
    is_planar: bool = False,
    format: Optional[Union[str, av.AudioFormat]] = None,
    rate: Optional[int] = None,
    to_mono: bool = False,
):
    kwargs = {}
    if dtype is not None:
        kwargs["sample_fmts"] = get_format(dtype, is_planar).name
    if format is not None:
        kwargs["sample_fmts"] = format.name if isinstance(format, av.AudioFormat) else format
    if rate is not None:
        kwargs["sample_rates"] = rate
    if to_mono:
        kwargs["channel_layouts"] = "mono"
    return filter.aformat(**kwargs)


__all__ = [
    "AudioCache",
    "AudioFormat",
    "AudioLayout",
    "ContainerFormat",
    "Decodec",
    "Encodec",
    "Filter",
    "Graph",
    "aformat",
    "audio_formats",
    "audio_layouts",
    "canonical_names",
    "clip",
    "container_formats",
    "decodecs",
    "encodecs",
    "extension_formats",
    "from_ndarray",
    "get_codecs",
    "get_dtype",
    "get_format",
    "load_url",
    "split_audio_frame",
    "standard_channel_layouts",
    "to_ndarray",
]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
from io import BytesIO
from typing import Any, Optional

import numpy as np

from audiolab.av.typing import Dtype


class Backend:
    def __init__(self, file: Any, sample_rate: int, dtype: Optional[Dtype] = None, format: str = "WAV"):
        self.file = file
        self.sample_rate = sample_rate
        self.dtype = None
        if dtype is not None:
            self.dtype = np.dtype(dtype)
        self.format = format

        self.is_closed = False
        atexit.register(self.close)

    def close(self):
        if not self.is_closed:
            if isinstance(self.file, BytesIO):
                self.file.seek(0)
            self.is_closed = True



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import wave
from typing import Any, Optional

import numpy as np

from audiolab.av.frame import clip
from audiolab.av.typing import Dtype
from audiolab.writer.backend.backend import Backend

_dtype_to_bytes = {"uint8": 1, "int16": 2, "int32": 4}


class Wave(Backend):
    def __init__(self, file: Any, sample_rate: int, dtype: Optional[Dtype] = None):
        super().__init__(file, sample_rate, dtype)
        self.wave = None
        self.num_channels = None

    def open(self):
        self.wave = wave.open(self.file, "w")
        self.wave.setframerate(self.sample_rate)
        self.wave.setnchannels(self.num_channels)
        sampwidth = _dtype_to_bytes[self.dtype.name]
        self.wave.setsampwidth(sampwidth)

    def write(self, frame: np.ndarray):
        if self.dtype is None:
            self.dtype = frame.dtype
        frame = np.atleast_2d(clip(frame, self.dtype))
        if self.num_channels is None:
            self.num_channels = frame.shape[0]
        if self.wave is None:
            self.open()
        self.wave.writeframes(frame.tobytes())

    def close(self):
        if self.wave is not None and not self.is_closed:
            self.wave.close()
            super().close()



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Dict, Tuple, Union

import av
import numpy as np


class BaseEnum(Enum):
    def __new__(cls, value):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __getattr__(self, attr):
        return getattr(self.value, attr)


class AudioFormatEnum(BaseEnum):
    pass


class AudioLayoutEnum(BaseEnum):
    pass


class CodecEnum(BaseEnum):
    pass


class ContainerFormatEnum(BaseEnum):
    pass


UINT32_MAX = np.iinfo(np.uint32).max
AudioFormat = Union[str, av.AudioFormat]
AudioFrame = Union[np.ndarray, av.AudioFrame, Tuple[np.ndarray, int]]
AudioLayout = Union[int, str, av.AudioLayout]
Codec = Union[str, av.Codec]
ContainerFormat = Union[str, av.ContainerFormat]
Dtype = Union[str, type, np.dtype]
Filter = Union[str, Tuple[str, str], Tuple[str, Dict[str, str]], Tuple[str, str, Dict[str, str]]]
Seconds = float



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, Set

import av

from audiolab.av.typing import ContainerFormatEnum
from audiolab.av.utils import get_template

"""
$ ffmpeg -formats
"""
container_formats: Dict[str, av.ContainerFormat] = {}
extension_formats: Dict[str, Set[str]] = defaultdict(set)
for name in av.formats_available:
    container_formats[name] = av.ContainerFormat(name)
    for extension in container_formats[name].extensions:
        extension_formats[extension].add(name)
ContainerFormat = ContainerFormatEnum("ContainerFormat", container_formats)


template = get_template("container")
for name, format in container_formats.items():
    getattr(ContainerFormat, name).__doc__ = template.render(format=format)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from base64 import b64encode
from io import BytesIO
from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np

from audiolab.av import (
    AudioCache,
    clip,
    from_ndarray,
    get_dtype,
    get_format,
    split_audio_frame,
    to_ndarray,
)
from audiolab.av.typing import Dtype
from audiolab.pipe import AudioPipe
from audiolab.reader import Reader, StreamReader, aformat, info, load_audio
from audiolab.writer import Writer, save_audio


def encode(
    audio: Union[str, Path, np.ndarray],
    rate: Optional[int] = None,
    dtype: Optional[Dtype] = None,
    to_mono: bool = False,
    make_wav: bool = True,
    format: str = "WAV",
) -> Tuple[str, int]:
    """
    Transform an audio to a PCM bytestring.

    Args:
        audio: The file path to an audio file or a numpy array.
        rate: The sample rate of the audio.
        dtype: The data type of the audio.
        to_mono: Whether to convert the audio to mono.
        make_wav: Whether to make the audio a WAV file.
        format: The format of the audio container.
    Returns:
        The audio as a PCM bytestring and the sample rate of the audio.
    """
    if isinstance(audio, (str, Path)):
        audio, rate = load_audio(audio, dtype=dtype, rate=rate, to_mono=to_mono)

    audio = clip(audio, np.int16)
    if make_wav:
        bytestream = BytesIO()
        save_audio(bytestream, audio, rate, format=format)
        audio = b64encode(bytestream.getvalue()).decode("ascii")
        audio = f"data:audio/{format};base64,{audio}"
    else:
        audio = np.ascontiguousarray(audio)
        audio = b64encode(audio).decode("ascii")
    return audio, rate


__all__ = [
    "AudioCache",
    "AudioPipe",
    "Reader",
    "StreamReader",
    "Writer",
    "aformat",
    "encode",
    "from_ndarray",
    "get_dtype",
    "get_format",
    "info",
    "load_audio",
    "save_audio",
    "split_audio_frame",
    "to_ndarray",
]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List

from av import filter
from av.option import OptionType

from audiolab.av.utils import get_template

"""
$ ffmpeg -filters
"""


class FilterManager:
    def __init__(self):
        self._filter_data: Dict[str, Dict[str, Any]] = {}
        self._initialized: bool = False

    def _generate_filter_data(self) -> None:
        for name in filter.filters_available:
            options = []
            _filter = filter.Filter(name)
            if _filter.options is not None:
                for opt in _filter.options:
                    try:
                        opt_type = opt.type
                    except ValueError:
                        opt_type = OptionType.STRING
                    options.append(
                        {
                            "name": opt.name,
                            "type": opt_type,
                            "default": opt.default,
                            "help": opt.help if opt.name != "temp" else "set temperature °C",
                        }
                    )
            self._filter_data[name] = {"name": _filter.name, "description": _filter.description, "options": options}

    def _create_filter_function(self, name: str):
        def filter_func(args=None, **kwargs):
            return (name, None if args is None else str(args), {k: str(v) for k, v in kwargs.items()})

        filter_func.__name__ = name
        return filter_func

    def _initialize_filters(self) -> None:
        if self._initialized:
            return

        self._generate_filter_data()
        for name in filter.filters_available:
            filter_func = self._create_filter_function(name)
            data = self._filter_data[name]
            filter_func.__doc__ = get_template("filter").render(
                name=data["name"], description=data["description"], options=data["options"]
            )
            globals()[name] = filter_func

        self._initialized = True

    def __getattr__(self, name: str) -> Callable:
        self._initialize_filters()
        return globals().get(name, None)

    @property
    def filters(self) -> List[str]:
        return filter.filters_available


_filter_manager = FilterManager()
filters = _filter_manager.filters


def __getattr__(name: str) -> Callable:
    return getattr(_filter_manager, name)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fractions import Fraction
from typing import Optional, Tuple

import av
import numpy as np

from audiolab.av.format import get_dtype
from audiolab.av.typing import AudioFormat, AudioLayout, Dtype
from audiolab.av.utils import get_logger

logger = get_logger(__name__)


def clip(ndarray: np.ndarray, dtype: Dtype) -> np.ndarray:
    src_dtype = ndarray.dtype
    dst_dtype = np.dtype(dtype)
    if src_dtype.kind != "f" and src_dtype == dst_dtype:
        return ndarray

    if src_dtype.kind == "f":
        min_value, max_value = ndarray.min(), ndarray.max()
        if min_value < -1.0 or max_value > 1.0:
            logger.warning("Cliping %s ndarray from: %g ~ %g to -1.0 ~ 1.0", src_dtype, min_value, max_value)
            ndarray = np.clip(ndarray, -1.0, 1.0)
    else:
        ndarray = ndarray.astype(np.float64)
        if src_dtype.kind == "u":
            ndarray = ndarray / np.iinfo(src_dtype).max * 2 - 1
        elif src_dtype.kind == "i":
            ndarray = ndarray / np.iinfo(src_dtype).max

    if dst_dtype.kind in ("u", "i"):
        max_value = np.float64(np.iinfo(dst_dtype).max)
        if dst_dtype.kind == "u":
            ndarray = (ndarray + 1) * 0.5 * max_value
        else:
            ndarray = ndarray * max_value
    return np.asarray(ndarray, dtype=dst_dtype)


def from_ndarray(
    ndarray: np.ndarray,
    format: AudioFormat,
    layout: AudioLayout,
    rate: int,
    pts: Optional[int] = None,
    time_base: Optional[Fraction] = None,
) -> av.AudioFrame:
    ndarray = np.atleast_2d(ndarray)
    if isinstance(format, str):
        format = av.AudioFormat(format)
    if format.is_packed:
        # [num_channels, num_samples] => [1, num_channels * num_samples]
        ndarray = ndarray.T.reshape(1, -1)
    if isinstance(layout, str):
        layout = av.AudioLayout(layout)

    dtype = get_dtype(format)
    ndarray = clip(ndarray, dtype)
    ndarray = np.ascontiguousarray(ndarray)
    frame = av.AudioFrame.from_ndarray(ndarray, format.name, layout)
    frame.rate = rate
    if pts is not None:
        frame.pts = pts
    if time_base is not None:
        frame.time_base = time_base
    return frame


def to_ndarray(frame: av.AudioFrame) -> np.ndarray:
    # packed: [num_channels, num_samples]
    # planar: [1, num_channels * num_samples]
    ndarray = frame.to_ndarray()
    if frame.format.is_packed:
        ndarray = ndarray.reshape(-1, frame.layout.nb_channels).T
    return ndarray


def split_audio_frame(frame: av.AudioFrame, offset: int) -> Tuple[av.AudioFrame, av.AudioFrame]:
    if offset <= 0:
        return None, frame
    # number of samples per channel
    if offset >= frame.samples:
        return frame, None

    ndarray = to_ndarray(frame)
    left, right = ndarray[:, :offset], ndarray[:, offset:]
    if frame.format.is_packed:
        left, right = left.T.reshape(1, -1), right.T.reshape(1, -1)
    left = av.AudioFrame.from_ndarray(left, frame.format.name, frame.layout)
    right = av.AudioFrame.from_ndarray(right, frame.format.name, frame.layout)
    left.rate, right.rate = frame.rate, frame.rate
    if frame.pts is not None:
        left.pts, right.pts = frame.pts, frame.pts + offset
    if frame.time_base is not None:
        left.time_base, right.time_base = frame.time_base, frame.time_base
    return left, right


def pad(frame: np.ndarray, frame_size: int, fill_value: float = 0) -> np.ndarray:
    pad_needed = frame_size - frame.shape[0 if frame.ndim == 1 else 1]
    if pad_needed <= 0:
        return frame
    if frame.ndim == 1:
        return np.pad(frame, (0, pad_needed), constant_values=fill_value)
    else:
        return np.pad(frame, ((0, 0), (0, pad_needed)), constant_values=fill_value)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, Set

import av
import numpy as np
from av import codecs_available
from av.codec.codec import UnknownCodecError

from audiolab.av.format import format_dtypes
from audiolab.av.typing import CodecEnum
from audiolab.av.utils import get_template

"""
$ ffmpeg -codecs
$ ffmpeg -decoders
$ ffmpeg -encoders
"""


class CodecManager:
    def __init__(self):
        self.canonical_names: Dict[str, Set[str]] = defaultdict(set)
        self.decodecs: Dict[str, av.Codec] = {}
        self.encodecs: Dict[str, av.Codec] = {}

        for codec in codecs_available:
            try:
                decoder_codec = av.Codec(codec)
                if decoder_codec.type != "audio":
                    continue
                if decoder_codec.audio_formats is not None:
                    canonical_name = decoder_codec.canonical_name
                    codec_name = decoder_codec.name
                    self.canonical_names[canonical_name].add(codec_name)
                    if codec_name not in self.decodecs:
                        self.decodecs[codec_name] = decoder_codec

                encoder_codec = av.Codec(codec, "w")
                if encoder_codec.audio_formats is not None:
                    canonical_name = encoder_codec.canonical_name
                    codec_name = encoder_codec.name
                    self.canonical_names[canonical_name].add(codec_name)
                    if codec_name not in self.encodecs:
                        self.encodecs[codec_name] = encoder_codec
            except UnknownCodecError:
                pass

        self.Decodec = CodecEnum("Decodec", self.decodecs)
        self.Encodec = CodecEnum("Encodec", self.encodecs)

        template = get_template("codec")
        for name, codec in self.decodecs.items():
            getattr(self.Decodec, name).__doc__ = template.render(codec=codec, format_dtypes=format_dtypes, np=np)
        for name, codec in self.encodecs.items():
            getattr(self.Encodec, name).__doc__ = template.render(codec=codec, format_dtypes=format_dtypes, np=np)


_codec_manager = CodecManager()
canonical_names = _codec_manager.canonical_names
decodecs = _codec_manager.decodecs
encodecs = _codec_manager.encodecs
Decodec = _codec_manager.Decodec
Encodec = _codec_manager.Encodec



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterator, List, Optional, Tuple

import numpy as np

from audiolab.av.frame import pad
from audiolab.av.typing import AudioFormat, Dtype, Filter
from audiolab.reader import Graph, aformat


class AudioPipe:
    def __init__(
        self,
        in_rate: int,
        filters: Optional[List[Filter]] = None,
        dtype: Optional[Dtype] = None,
        is_planar: bool = False,
        format: Optional[AudioFormat] = None,
        out_rate: Optional[int] = None,
        to_mono: bool = False,
        frame_size: Optional[int] = 1024,
        fill_value: Optional[float] = None,
        always_2d: bool = True,
    ):
        self.in_rate = in_rate
        self.graph = None
        if not all([dtype is None, format is None, out_rate is None, to_mono is None]):
            filters = filters or []
            filters.append(aformat(dtype, is_planar, format, out_rate, to_mono))
        self.filters = filters
        self.frame_size = frame_size
        self.fill_value = fill_value
        self.always_2d = always_2d

    def push(self, frame: np.ndarray):
        if self.graph is None:
            self.graph = Graph(
                rate=self.in_rate,
                dtype=frame.dtype,
                channels=frame.shape[0],
                filters=self.filters,
                frame_size=self.frame_size,
                return_ndarray=True,
            )
        self.graph.push(frame)

    def pull(self, partial: bool = False) -> Iterator[Tuple[np.ndarray, int]]:
        for frame, rate in self.graph.pull(partial=partial):
            if self.fill_value is not None:
                frame = pad(frame, self.frame_size, self.fill_value)
            yield frame if self.always_2d else frame.squeeze(), rate



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from audiolab.writer.backend.backend import Backend
from audiolab.writer.backend.pyav import PyAV as pyav
from audiolab.writer.backend.soundfile import SoundFile as soundfile
from audiolab.writer.backend.wave import Wave as wave

__all__ = ["Backend", "pyav", "soundfile", "wave"]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from typing import Iterator, List, Optional

import av
from av import AudioCodecContext

from audiolab.av import aformat
from audiolab.av.graph import Graph
from audiolab.av.typing import AudioFormat, AudioFrame, Dtype, Filter


class StreamReader:
    def __init__(
        self,
        filters: Optional[List[Filter]] = None,
        dtype: Optional[Dtype] = None,
        is_planar: bool = False,
        format: Optional[AudioFormat] = None,
        rate: Optional[int] = None,
        to_mono: bool = False,
        frame_size: Optional[int] = 1024,
        return_ndarray: bool = True,
        always_2d: bool = True,
    ):
        """
        Create a StreamReader object.

        Args:
            filters: The filters to apply to the audio stream.
            dtype: The data type of the output audio frames.
            is_planar: Whether the output audio frames are planar.
            format: The format of the output audio frames.
            rate: The sample rate of the output audio frames.
            to_mono: Whether to convert the output audio frames to mono.
            frame_size: The frame size of the audio frames.
            return_ndarray: Whether to return the audio frames as ndarrays.
        """
        self._codec_context = None
        self._graph = None
        self.bytes_io = BytesIO()
        self.bytes_per_decode_attempt = 0
        if not all([dtype is None, format is None, rate is None, to_mono is None]):
            filters = filters or []
            filters.append(aformat(dtype, is_planar, format, rate, to_mono))
        self.filters = filters
        self.frame_size = frame_size
        self.return_ndarray = return_ndarray
        self.always_2d = always_2d
        self.offset = None
        self.packet = None

    @property
    def codec_context(self) -> Optional[AudioCodecContext]:
        if self._codec_context is None:
            if self.packet is None:
                return None
            self._codec_context = self.packet.stream.codec_context
            assert self._codec_context.frame_size in (0, 1)
        return self._codec_context

    @property
    def graph(self) -> Optional[Graph]:
        if self._graph is None:
            if self.packet is None:
                return None
            self._graph = Graph(
                self.packet.stream,
                filters=self.filters,
                frame_size=self.frame_size,
                return_ndarray=self.return_ndarray,
                always_2d=self.always_2d,
            )
        return self._graph

    def push(self, frame: bytes):
        self.bytes_io.write(frame)
        self.bytes_per_decode_attempt += len(frame)

    def pull(self, partial: bool = False) -> Iterator[AudioFrame]:
        if partial or self.bytes_per_decode_attempt * 2 >= self.frame_size:
            self.bytes_per_decode_attempt = 0
            try:
                self.bytes_io.seek(0)
                container = av.open(self.bytes_io, metadata_encoding="latin1")
                for packet in container.demux():
                    self.packet = packet
                    if self.packet.pts is None and not partial:
                        continue
                    # o: current frame
                    # pts: self.offset, frame.pts, packet.pts
                    # +---+---+---+---+---+
                    # | x | x | x | o |   |
                    # +---+---+---+---+---+
                    #             ↑
                    #             pts
                    if self.offset is not None and (self.packet.pts is None or self.offset > self.packet.pts):
                        continue
                    for frame in self.codec_context.decode(packet):
                        self.offset = frame.pts + int(frame.samples / packet.stream.rate / packet.stream.time_base)
                        self.graph.push(frame)
                        yield from self.graph.pull()
                    yield from self.graph.pull(partial=partial)
            except (av.EOFError, av.InvalidDataError, av.OSError, av.PermissionError):
                pass

    def reset(self):
        self._codec_context = None
        self._graph = None
        self.bytes_io = BytesIO()
        self.bytes_per_decode_attempt = 0
        self.offset = None
        self.packet = None



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import av

from audiolab.av.typing import AudioLayoutEnum

"""
$ ffmpeg -layouts
"""
standard_channel_layouts = {
    0: ["downmix"],
    1: ["mono"],
    2: ["stereo"],
    3: ["2.1", "3.0", "3.0(back)"],
    4: ["4.0", "quad", "quad(side)", "3.1"],
    5: ["5.0", "5.0(side)", "4.1"],
    6: ["5.1", "6.0", "6.0(front)", "hexagonal", "5.1(side)", "3.1.2"],
    7: ["7.0", "7.0(front)", "6.1", "6.1(back)", "6.1(front)"],
    8: ["7.1", "7.1(wide)", "7.1(wide-side)", "cube", "octagonal", "5.1.2"],
    10: ["5.1.4", "7.1.2"],
    12: ["7.1.4", "7.2.3"],
    14: ["9.1.4"],
    16: ["hexadecagonal"],
    24: ["22.2"],
}

audio_layouts: Dict[str, av.AudioLayout] = {
    name: av.AudioLayout(name) for layouts in standard_channel_layouts.values() for name in layouts
}
AudioLayout = AudioLayoutEnum("AudioLayout", audio_layouts)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property
from typing import Any, Optional

import numpy as np
import soundfile as sf

from audiolab.av.frame import clip
from audiolab.av.typing import Dtype
from audiolab.writer.backend.backend import Backend

_dtype_to_subtype = {"int16": "PCM_16", "int32": "PCM_32", "float32": "FLOAT", "float64": "DOUBLE"}


class SoundFile(Backend):
    def __init__(self, file: Any, sample_rate: int, dtype: Optional[Dtype] = None, format: str = "WAV"):
        super().__init__(file, sample_rate, dtype, format)
        self.sf = None
        self.num_channels = None

    @cached_property
    def subtype(self) -> str:
        if self.dtype is None:
            return sf.default_subtype(self.format)
        subtype = _dtype_to_subtype[self.dtype.name]
        # assert subtype in sf.available_subtypes(self.format)
        assert sf.check_format(self.format, subtype)
        return subtype

    def open(self):
        self.sf = sf.SoundFile(self.file, "w", self.sample_rate, self.num_channels, self.subtype, format=self.format)

    def write(self, frame: np.ndarray):
        if self.dtype is None:
            self.dtype = frame.dtype
        frame = np.atleast_2d(clip(frame, self.dtype))
        if self.num_channels is None:
            self.num_channels = frame.shape[0]
        if self.sf is None:
            self.open()
        # (num_channels, num_samples) => (num_samples, num_channels)
        self.sf.write(frame.T)

    def close(self):
        if self.sf is not None and not self.is_closed:
            self.sf.close()
            super().close()



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
from fractions import Fraction
from typing import List, Optional

import av
import numpy as np
from av import filter

from audiolab.av.format import get_format
from audiolab.av.frame import from_ndarray, to_ndarray
from audiolab.av.layout import standard_channel_layouts
from audiolab.av.typing import UINT32_MAX, AudioFormat, AudioFrame, AudioLayout, Dtype, Filter


class Graph(filter.Graph):
    def __init__(
        self,
        template: Optional[av.AudioStream] = None,
        rate: Optional[int] = None,
        dtype: Optional[Dtype] = None,
        is_planar: bool = False,
        format: Optional[AudioFormat] = None,
        layout: Optional[AudioLayout] = None,
        channels: Optional[int] = None,
        time_base: Optional[Fraction] = None,
        filters: Optional[List[Filter]] = None,
        frame_size: Optional[int] = None,
        return_ndarray: bool = True,
    ):
        if template is not None:
            rate = template.sample_rate if rate is None else rate
            format = template.format if format is None else format
            layout = template.layout.name if layout is None else layout
            channels = template.channels if channels is None else channels
            time_base = template.time_base if time_base is None else time_base
        format = get_format(dtype, is_planar) if format is None else format
        format = format.name if isinstance(format, av.AudioFormat) else format
        time_base = Fraction(1, rate) if time_base is None else time_base
        if layout is None:
            layout = standard_channel_layouts[channels][0]
        abuffer = super().add_abuffer(None, rate, format, layout, channels, time_base=time_base)

        nodes = [abuffer]
        if filters is not None:
            for _filter in filters:
                name, args, kwargs = (
                    (_filter, None, {})
                    if isinstance(_filter, str)
                    else ((*_filter, {}) if len(_filter) == 2 else _filter)
                )
                nodes.append(super().add(name, args, **kwargs))
        nodes.append(super().add("abuffersink"))
        super().link_nodes(*nodes).configure()

        self.frame_size = None
        if frame_size is not None and frame_size > 0:
            self.frame_size = min(frame_size, UINT32_MAX)
            super().set_audio_frame_size(self.frame_size)

        self.rate = rate
        self.format = format
        self.layout = layout
        self.return_ndarray = return_ndarray

    def push(self, frame: AudioFrame):
        if isinstance(frame, tuple):
            frame, rate = frame
            assert rate == self.rate
        if isinstance(frame, np.ndarray):
            frame = from_ndarray(frame, self.format, self.layout, self.rate)
        super().push(frame)

    def pull(self, partial: bool = False, return_ndarray: Optional[bool] = None) -> AudioFrame:
        if partial:
            super().push(None)
        while True:
            try:
                frame = super().pull()
                if return_ndarray is None:
                    return_ndarray = self.return_ndarray
                yield (to_ndarray(frame), frame.rate) if return_ndarray else frame
            except av.EOFError:
                break
            except av.FFmpegError as e:
                if e.errno != errno.EAGAIN:
                    raise
                break
